/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
#ifdef QUANT
#include "quant.h"
#endif

double **quant = NULL;

static int setupGenDrp (int, char **, char *, int);

void  initLodSc (char **);
void  procLodScPars (int, char **, char *);
void  chkPedFil (void);
double **setupQPed (void);
void  setPed (double **);
int   setSwitches (void);
void  setObsData (void);
void  setQObsData (double **);
void  setMaps (int);
void  allocInherit (void);

void  dbgGeno (void);
void  dbgXXGG (void);
void  dbgGGGeno (int);
void  dbgProb (double **);

/**
 * Calls functions to do the setup steps for the Morgan programs
 * which use a Monte Carlo method.
 * This includes setting up parameters and data structures for the
 * marker and trait model, 
 * reading the pedigree file and setting up the pedigree structure, 
 * storing observed marker and trait data for pedigree members; 
 * and setting initial values for inheritance indicators and 
 * Founder Gene Labels prior to burn-in and the main MCMC iterations.
 */
int setupGenDrp (int argc, char **argv, char *ppn, int nh)
{
   int   nSwtch;

   /* initialize exit procedure, print headings. 
    * fncs:  atexit(), print_head() 
    * dirs:  Stuff 
    */
   initLodSc (argv);

   /* process parameter statements, set up globals resulting from these.
    * fncs:  proc_pars(), prntStatusPP(), proc_auto(), proc_fils(), 
    *        proc_seeds(), prntStatusCP() 
    * vars:  globals in globs.h
    * dirs:  Pars, Stuff
    */
   procLodScPars (argc, argv, ppn);

   /* check, and massage if necessary, pedigree file.
    * fncs:  chk_pedfil(), prntStatusCPF(), mak_opedfil()
    * vars:  indKeys[PedSize], compntList[nCompnt]
    * dirs:  PedComp, Stuff
    */
   chkPedFil ();

   /* read in pedigree data and set up pedigree structure.
    * fncs:  alloc_ped(), open_ped(), input_ped(), count_founders() 
    * vars:  int data[PedSize][NColData], Ind nghd[PedSize+1]
    *        double quant[NTraits][PedSize], Ind nghd[].g_dat[2*(NMarks+1)]
    * dirs:  Nghds, Quant, Stuff
    */
#ifdef QUANT
   quant = setupQPed ();
#endif
   setPed (quant);

   nSwtch = setSwitches ();

#if 0
   /* set observed marker and trait data for pedigree members.
    * fncs:  set_markdat(), set_traitdat(), prnt_traitdat(), set_qdat()
    * vars:  Ind nghd[].g_dat[2*(NMarks+1)],
    *        Ind nghd[].d_dat[2*(NMarks+1)] 
    * dirs:  Markers
    */
   setObsData ();
#endif
#ifdef QUANT
   setQObsData (quant);
#endif

   /* set up chromosome map for markers and trait. */
   /* fncs:  alloc_chromo(), set_chromosomes()
    * vars:  int chromo[nh+1][NMarks+1], revchro[nh+1][NMarks+1],
    *        double xecomo[2][nh+1][NMarks+1]
    * dirs:  Markers
    */
   setMaps (nh);

   /* set up data structures for simulation of inheritance and FGL. 
    * fncs:  alloc_xxgg(), allocate_loci()
    * vars:  int xx0[NMarks+1][PedSize], xx1[NMarks+1][PedSize] 
    *        int gg0[NMarks+1][PedSize], gg1[NMarks+1][PedSize]
    *        Locus loci[NMarks+1].*  
    * dirs:  Sample
    */
   allocInherit ();

   /* initialize seeds for rans().
    * vars:  pSeedC, pSeedT   
    * dirs:  Rans
    */
   setSeeds (SAMPL);

   return nSwtch;
}

/* ........................................................................ */

/* the functions below are to be moved to appropriate directories  */

/**
 * Initialize exit procedure;  print header.
 */
void initLodSc (char **argv)
{
   atexit (print_tail);

   print_head (*argv);

   return;
}

/**
 * Process the parameter file and set corresponding globals
 * (these start with capital letters, and are found in globs.h)
 */
void procLodScPars (int argc, char **argv, char *ppn)
{

   /* parse input parameter statements  */
   if (!proc_pars (argc, argv, ppn))
      exit (EXIT_FAILURE);

   prntStatusPP ();

   /* check statements for completeness and consistency.
    * Set up global parameters and data structures     */
   if (!proc_auto ())
      exit (EXIT_FAILURE);

   /* set up pedigree file parameters */
   if (!proc_fils ())
      exit (EXIT_FAILURE);

   /* set up seed parameters */
   if (!proc_seeds ())
      exit (EXIT_FAILURE);

   prntStatusCP ();

   return;
}

/**
 * read and check pedigree file;  write new one if needed.
 */
void chkPedFil (void)
{
   /* check pedigree file for errors */
   if (!chk_pedfil ())
      give_up (ERR_qit);

   prntStatusCPF ();

   /* write new pedigree file, if changes to file */
   if (!mak_opedfil ())
      give_up (ERR_qit);

   return;
}

/**
 * Read and store pedigree data.
 */
void setPed (double **qdata)
{
   int   n;
   FILE *pedF;

   /* allocate structures for pedgree data:  int data[PedSize][NColData]
    *                                        Ind nghd[PedSize + 1]  */
   alloc_ped ();

   /* open for reading */
   if (!(pedF = open_ped ()))
      give_up (ERR_qit);

   /* read in pedigree and trait data:       data[][]  */
   input_ped (pedF, qdata);

   fclose (pedF);

   /* set up neighborhood and pedigree components data structures */
   nghds ();

   /* do an extra check on no. founders */
   if ((n = count_founders ()) != nFound)
      give_up_where (WHERE, ERN_cfd, nFound, n);

   /* set up lists of members in each component and also 
      information about each component */
   if ((n_compnts = components ()) != nCompnt)
      give_up_where (WHERE, ERN_ccp, nCompnt, n_compnts);

   return;
}

/**
 * set and print the switches for each component.
 */
int setSwitches (void)
{
   int   nsw;

   nsw = make_switches ();
   prnt_switches (nsw);

   return nsw;
}

/**
 *  set observed marker and trait genotypic data for pedigree members.
 */
void setObsData (void)
{
   /* set marker data, and trait data if genotypic, for observed individuals: 
    *                                        nghd[].g_dat[2*(NMarks+1)]  */
   set_markdat (NMarks, nCompnt);

   /* if trait data is discrete, set trait data from pedigree data   
    *                                        nghd[].d_dat[2*(NMarks+1)]  */
   if (TraitTyp != UNDEF)
   {
      if (!strcmp (ProgNam, "lm_auto") || !strcmp (ProgNam, "lm_ibdtests"))
      {
         set_traitdat (pedsize, n_compnts, 1);
         prnt_traitdat (PedSize, nCompnt);
      }
      else if (AffTrIn)
         set_traitdat (pedsize, n_compnts, 1);

      if (!strcmp (ProgNam, "lm_pval") || !strcmp (ProgNam, "lm_fuzzp"))
         set_affind (nCompnt, 1);       /* sets only affecteds  */
      if (!strcmp (ProgNam, "lm_ibdtests"))
         set_affind (nCompnt, 0);       /* sets only affecteds */
   }
   if (DebugMain)
      dbgGeno ();

   return;
}

/**
 * Setup structures for recombination, inheritance indicators, genes,
 * and loci
 */
void setMaps (int nh)
{
   int   sx;

   /* allocate and set arrays of marker and trait positions 
    * and recombination fractions  (for sampler setup, trait is unlinked).  
    *                                        int chromo[nh+1][NMarks+1]    
    *                                        revchro[nh+1][NMarks+1]      
    *                                        double xecomo[2][nh+1][NMarks+1] */
   alloc_chromo (NMarks, nh);
   for (sx = FEM; sx <= MAL; sx++)
      set_chromosomes (NMarks, Marek[sx], nh, &Loc0, &Rec0[sx], sx);

   return;
}

/**
 * Sets up data structures for locus and meiosis sampling.
 */
void allocInherit (void)
{

   /* allocate arrays xx0, xx1, for meiosis indicators                   
    *                                        int xx0 [NMarks+1][PedSize]
    *                                            xx1 [NMarks+1][PedSize]  
    * allocate arrays gg0, gg1 for genes                                 
    *                                            gg0 [NMarks+1][PedSize]  
    *                                            gg1 [NMarks+1][PedSize]  */
   alloc_xxgg (NMarks, PedSize);

   if (!strcmp (ProgNam, "ibddrop"))
      return;

   /* allocate Locus structure                         
    *                                        struct Locus loci[NMarks+1]  */
   if (TraitTyp == QUANTITATIVE)
      allocate_loci (NMarks + 1, NAll, Qq, 4);
   else if (PedDatTyp == DISCR_LIAB)
      allocate_loci (NMarks + 1, NAll, Qq, 3);
   else
      allocate_loci (NMarks + 1, NAll, Qq, 1);

   return;
}

/* ........................................................................ */

void dbgGeno (void)
{
   int   i, j;

   for (i = 0; i < PedSize; ++i)
      if (nghd[i].g_dat)
      {
         fprintf (stderr, "   ind = %*s ", widPNam, nghd[i].self);
         for (j = 0; j <= NMarks; ++j)
            fprintf (stderr, " %d %d", nghd[i].g_dat[2 * j],
                     nghd[i].g_dat[2 * j + 1]);
         fprintf (stderr, "\n");
      }
   fprintf (stderr, "\n");
   fflush (stderr);

   return;
}

void dbgXXGG (void)
{
   int   i, j;

   for (j = 0; j < NMarks + 1; ++j)
   {
      fprintf (stderr, "Locus %d position %d set-up\n", j, chromo[0][j]);
      for (i = 0; i < PedSize; ++i)
      {
         fprintf (stderr, " %3d %*s", i, widPNam, nghd[i].self);
         fprintf (stderr, "  %*s  %*s", widPNam, nghd[i].pa->self,
                  widPNam, nghd[i].ma->self);
         fprintf (stderr, "   %2d  %2d   %2d  %2d\n",
                  xx1[j][i], xx0[j][i], gg1[j][i], gg0[j][i]);
      }
      fprintf (stderr, "\n");
      fflush (stderr);
   }
   return;
}

void dbgGGGeno (int j)
{
   int   i, ii;

   for (i = 0; i < NObs; ++i)
   {
      ii = ObsIx[i];
      fprintf (stderr, "   %2d %*s %2d   %2d %2d   %2d %2d",
               i, widPNam, nghd[ii].self, ii, gg1[j][ii], gg0[j][ii],
               nghd[ii].g_dat[2 * j], nghd[ii].g_dat[2 * j + 1]);
   }
   fprintf (stderr, "\n");
   fflush (stderr);
   return;
}
void dbgProb (double **p)
{
   int   j, c;

   for (j = 0; j < NMarks + 1; ++j)
   {
      fprintf (stderr, " Set genes at locus %d:\n", j);
      dbgGGGeno (j);
   }
   for (j = 0; j < NMarks + 1; ++j)
   {
      fprintf (stderr, " Scored prob at locus %d; giving", j);
      fflush (stderr);
      for (c = 0; c < nCompnt; ++c)
         fprintf (stderr, "   Component %d:  %7.4f\n", c, p[c][j]);
   }
   fprintf (stderr, "\n");
   fflush (stderr);
   return;
}
