/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/************************************************************************
   *                                                                    *
   *      Define a Trait structure                                      *
   *                                                                    *
   *     Traits have names -- Cholesterol, Alzheimers, D1S479           *
   *     Traits have types-- continuous, discrete, genotypic            *
   *     Traits have list of loci contributing to the trait (often 1)   *
   *     Traits have associated penetrance functions which define the   *
   *         probability of the phenotype given
   *     the genotype at each contributing locus
   *     the  age, sex, and other observed covariates of an individual
   *
   *     Covariates have names: sex, age-class, age, blood-pressure
   *      they may be discrete (few levels) or "continuous" multi-level
   *      discrete ones have some number of levels.
   *     There is a Covlist of total covariates input in analysis -- 
   *         and recorded for observed individuals 
   *     There is Covlist of the covariates contributing to a given trait.
   *
   *     What is penetrance
   *          -- a function of the trait,requiring
   *            the individual  (genotype, covariate values, etc)
   *            the locus list  (to know which genotypes impact trait)
   *            the covariate list (to know which covariates .... )
   *     penetrance is a function that is a property of a trait, and
   *       also a function that includes the Trait (and hence
   *       associated Loclist and Covlist etc) as an argument.
   *
   *     EAT  20:02 20/02/2002  -- the ultimate palindromic date/time   *
   *                                                                    *
*************************************************************************/

#ifndef TRAITS_H
#define TRAITS_H

#include "markers.h"

/**
 * @addtogroup markers
 * @{
 */

/* need to separate traits and markers defined trait types  */
#if 0                             /* now in limdefs.h */
#define MARKER_GENO -1 
#define GENOTYPIC    0  
#define DISCRETE     1
#define ORDINAL      2
#define QUANTITATIVE 3
#endif

/**********************************************************************
Defaults for compatibility with MORGAN:
If nothing is specified:
Trait 0 corresponds to locus 0 (the "trait locus"), and has type
      GENOTYPIC (current default), 
      DISCRETE (if "set trait data discrete"),
   or CONTINUOUS (if "set trait data quantitative")
Traits 1,2,..., nloc 
      correspond to marker loci 1,2,...,nloc
      and each has type MARKER-GENO.

Markers in the parameter and map statements correspond to markers in the
marker data file or set marker data statements, and are ordered along a
chromosome.
The trait setup will specify which markers are to be used.
A marker trait (type -1) will have a locus index, which is the column in the
marker data file, which will become the locus_index  which is the map index of
the markers USED in the analysis
(which may or may not be in the same order as trait indices.)

Given the marker traits to be used, the marker data array will be set to
only those markers ordered along the chromosome.
nloc (the number of marker loci) will be reset
Marek (male and female) will be reset to map for those loci.

The marker data arrays will NOT be changed from the current spec.
What about map trait statements in lm_auto ?? -- need to reset loc0 and rec0 ?
************************************************************************/


struct Trait {
   char self[MAXNAMLEN+1];            /**< char name for trait */
   int trait_type;                    /**< contin, discrete or genotype */
   int index;                         /**< number the traits   */
   int locus;                         /**< simple case 1-1 corresp */
   struct _Censor  *censor;           /**< censoring status/values */
   struct _Loclist *loci;             /**< loci contrib to trait */
   struct _Covlist *covar;            /**< covariates contrib to trait */
   void (*penetrance) (double *,  const struct Trait *,
                    const Ind *, const struct _Loclist *, double);  
}; 

typedef struct _Loclist {
   struct Locus *current;
   struct _Loclist *next;
} Loclist;

typedef struct _Covariate {
   char self[MAXNAMLEN+1];           /**< char name for covariate  */
   int covar_type;                   /**< continuous, discrete     */
   int levels;                       /**< number of classes/levels */
   struct Censor  *censor;           /**< censoring status/values  */
} Covariate;

typedef struct _Censor {
   int left_censor;                  /* a trait or covariate may be censored */
   int right_censor;                 /*   to the left or the right True False */
   double left_value;                /* if true -- need left/right censoring */
   double right_value;               /*   values                             */
} Censor;

typedef struct _Covlist {
  struct Covariate *current;
  struct Covlist *next;
} Covlist;

GLOBAL struct Trait *traits;

/* folowing routines in set_traits.c */
int  set_eval_points (int , int *, int *,
        double *, double *, double * , double *);   /*lm_markrs rtn */


int set_traits (int, int *);                   /* returns total traits */
void default_traits (int);             
void read_traits (int, FILE*);        
int print_traits (int, int *);                 /* returns new marker count */
void free_traits (void);                       /* frees traits */
void flag_loci (int, int, int*);

void recess2(double *, const struct Trait *,
                    const Ind *, const struct _Loclist * , double);
                                /* trait penetrance -- under development */

/* following routines in set_markdat.c:    EAT Mar  2002
 * moved to traits.h                       EAT Mar  2002
 * moved to markers.h                      MAJ May  2003
 * void unset_recval(double **, double **, double **, double **);
 * void reset_recval (int, double **, double **, double **, double **);
 * void set_markmap(int , int , int *);
 * void reset_traitmap (int, int, int*);
 * void set_markdat   (int ***, int, int, int *);  
 */

/** @} */

#endif
