/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/* 
 NAME

   alloc_char2d, alloc_char3d, alloc_int2d_two, alloc_int3d_two,
   alloc_dbl2d_two, alloc_dbl3d_two, 
   free_char2d, free_int2d, free_dbl2d, free_dbl3d

 SYNOPSIS

   #include "twoqtl.h"

   char **alloc_char2d (int dim1, int dim2);
   char ***alloc_char3d (int dim1, int dim2, int dim3);

   int **alloc_int2d_two (int dim1, int dim2);
   int ***alloc_int3d_two (int dim1, int dim2, int dim3);
   double **alloc_dbl2d_two (int dim1, int dim2);
   double ***alloc_dbl3d_two (int dim1, int dim2, int dim3);

   void   free_char2d (void *);
   void   free_int2d (void *);
   void   free_dbl2d (void *);
   void   free_dbl3d (void *);

 DESCRIPTION
   
   alloc functions allocate 2d or 3d arrays in consecutive blocks of memory
   with fewer mallocs (differing from those in alloc_array.c in Stuff). 
 
   free functions free arrays with proper argument 
   without assigning null to pointer (again differing from those in 
   alloc_array.c, because void ** is not valid c type).

 HISTORY

   updated from alloc_array.c in Stuff        Feb 14 2005
   any possibility with merging into alloc_array ??

 Author 
   Written by Yun Ju Sung

 */

#include "twoqtl.h"

char **alloc_char2d (int dim1, int dim2)
{
   int   i;

   /* allocate */
   char **array2d = (char **) malloc (dim1 * sizeof (char *));
   alloc_chk (WHERE, array2d);

   array2d[0] = (char *) malloc (dim1 * dim2 * sizeof (char));
   alloc_chk (WHERE, array2d[0]);
   for (i = 1; i < dim1; ++i)
      array2d[i] = array2d[i - 1] + dim2;

   /* initialize */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (char));

   return array2d;
}

char ***alloc_char3d (int dim1, int dim2, int dim3)
{
   int   i, j;
   char ***array3d;
   char *tmp;

   /* allocate */
   array3d = (char ***) malloc (dim1 * sizeof (char **));
   alloc_chk (WHERE, array3d);

   array3d[0] = (char **) malloc (dim1 * dim2 * sizeof (char *));
   alloc_chk (WHERE, array3d[0]);
   for (i = 1; i < dim1; ++i)
      array3d[i] = array3d[i - 1] + dim2;

   array3d[0][0] = (char *) malloc (dim1 * dim2 * dim3 * sizeof (char));
   alloc_chk (WHERE, array3d[0][0]);

   tmp = array3d[0][0];
   for (i = 0; i < dim1; ++i)
      for (j = 0; j < dim2; ++j)
      {
         array3d[i][j] = tmp;
         tmp += dim3;
      }

   /* initialize */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (char));

   return array3d;
}

int **alloc_int2d_two (int dim1, int dim2)
{
   int   i;
   int **array2d;

   /* allocate */
   array2d = (int **) malloc (dim1 * sizeof (int *));
   alloc_chk (WHERE, array2d);

   array2d[0] = (int *) malloc (dim1 * dim2 * sizeof (int));
   alloc_chk (WHERE, array2d[0]);
   for (i = 1; i < dim1; ++i)
      array2d[i] = array2d[i - 1] + dim2;

   /* initialize */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (int));

   return array2d;
}

int ***alloc_int3d_two (int dim1, int dim2, int dim3)
{
   int   i, j;
   int ***array3d;
   int  *tmp;

   /* allocate */
   array3d = (int ***) malloc (dim1 * sizeof (int **));
   alloc_chk (WHERE, array3d);

   array3d[0] = (int **) malloc (dim1 * dim2 * sizeof (int *));
   alloc_chk (WHERE, array3d[0]);
   for (i = 1; i < dim1; ++i)
      array3d[i] = array3d[i - 1] + dim2;

   array3d[0][0] = (int *) malloc (dim1 * dim2 * dim3 * sizeof (int));
   alloc_chk (WHERE, array3d[0][0]);

   tmp = array3d[0][0];
   for (i = 0; i < dim1; ++i)
      for (j = 0; j < dim2; ++j)
      {
         array3d[i][j] = tmp;
         tmp += dim3;
      }

   /* initialize */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (int));

   return array3d;
}

double **alloc_dbl2d_two (int dim1, int dim2)
{
   int   i;
   double **array2d;

   /* allocate */
   array2d = (double **) malloc (dim1 * sizeof (double *));
   alloc_chk (WHERE, array2d);

   array2d[0] = (double *) malloc (dim1 * dim2 * sizeof (double));
   alloc_chk (WHERE, array2d[0]);
   for (i = 1; i < dim1; ++i)
      array2d[i] = array2d[i - 1] + dim2;

   /* initialize */
   memset (array2d[0], 0.0, dim1 * dim2 * sizeof (double));

   return array2d;
}

double ***alloc_dbl3d_two (int dim1, int dim2, int dim3)
{
   int   i, j;
   double ***array3d;
   double *tmp;

   /* allocate */
   array3d = (double ***) malloc (dim1 * sizeof (double **));
   alloc_chk (WHERE, array3d);

   array3d[0] = (double **) malloc (dim1 * dim2 * sizeof (double *));
   alloc_chk (WHERE, array3d[0]);
   for (i = 1; i < dim1; ++i)
      array3d[i] = array3d[i - 1] + dim2;

   array3d[0][0] = (double *) malloc (dim1 * dim2 * dim3 * sizeof (double));
   alloc_chk (WHERE, array3d[0][0]);

   tmp = array3d[0][0];
   for (i = 0; i < dim1; ++i)
      for (j = 0; j < dim2; ++j)
      {
         array3d[i][j] = tmp;
         tmp += dim3;
      }

   /* initialize */
   memset (array3d[0][0], 0.0, dim1 * dim2 * dim3 * sizeof (double));

   return array3d;
}

void free_char2d (void *foo)
{
   char **array = foo;
   free (*array);
   free (array);
   return;
}

void free_int2d (void *foo)
{
   int **array = foo;
   free (*array);
   free (array);
   return;
}

void free_dbl2d (void *foo)
{
   double **array = foo;
   free (*array);
   free (array);
   return;
}

void free_dbl3d (void *foo)
{
   double ***array = foo;
   free (**array);
   free (*array);
   free (array);
   return;
}
